from __future__ import unicode_literals

from django.core.files.uploadedfile import SimpleUploadedFile
from django.utils import unittest

from mock import patch

from ..forms import UploadForm
from ..utils import serialize_upload
from .base import MockStorage


class UploadFormTestCase(unittest.TestCase):
    """Default upload form logic."""

    def test_basic_stash(self):
        """
        Stash file using the passed storage.
        """
        data = {}
        files = {'upload': SimpleUploadedFile('test.jpg', content=b'X')}
        form = UploadForm(data=data, files=files)
        storage = MockStorage()
        result = form.stash(storage, '/upload/default/')
        expected = {
            'filename': 'test.jpg',
            'url': '/uploads/test.jpg',
            'stored': serialize_upload('test.jpg', storage, '/upload/default/'),
        }
        self.assertEqual(result, expected)

    def test_invalid_stash(self):
        """
        Stash result should be empty when form is not valid.
        """
        form = UploadForm(data={}, files={})
        storage = MockStorage()
        result = form.stash(storage, '/upload/default/')
        self.assertEqual(result, {})

    def test_stash_url(self):
        """
        Stash result should contain url if implement by
        the storage class otherwise None.
        """
        data = {}
        files = {'upload': SimpleUploadedFile('test.jpg', content=b'X')}
        form = UploadForm(data=data, files=files)
        storage = MockStorage()
        with patch.object(storage, 'url') as mock_url:
            mock_url.side_effect = NotImplementedError
            result = form.stash(storage, '/upload/default/')
        expected = {
            'filename': 'test.jpg',
            'url': None,
            'stored': serialize_upload('test.jpg', storage, '/upload/default/'),
        }
        self.assertEqual(result, expected)


"""Form logic for background saving uploaded files."""

from django import forms

from .utils import serialize_upload


class UploadForm(forms.Form):
    """Base class for accepting file upload."""

    upload = forms.FileField()

    def stash(self, storage, url):
        """Stores the uploaded file in a temporary storage location."""
        result = {}
        if self.is_valid():
            upload = self.cleaned_data['upload']
            name = storage.save(upload.name, upload)
            result['filename'] = name
            try:
                result['url'] = storage.url(name)
            except NotImplementedError:
                result['url'] = None
            result['stored'] = serialize_upload(name, storage, url)
        return result


from __future__ import unicode_literals

import os

from django.core import signing
from django.core.exceptions import ImproperlyConfigured
from django.core.files.storage import get_storage_class
from django.utils.functional import LazyObject


def serialize_upload(name, storage, url):
    """
    Serialize uploaded file by name and storage. Namespaced by the upload url.
    """
    if isinstance(storage, LazyObject):
        # Unwrap lazy storage class
        storage._setup()
        cls = storage._wrapped.__class__
    else:
        cls = storage.__class__
    return signing.dumps({
        'name': name,
        'storage': '%s.%s' % (cls.__module__, cls.__name__)
    }, salt=url)


def deserialize_upload(value, url):
    """
    Restore file and name and storage from serialized value and the upload url.
    """
    result = {'name': None, 'storage': None}
    try:
        result = signing.loads(value, salt=url)
    except signing.BadSignature:
        # TODO: Log invalid signature
        pass
    else:
        try:
            result['storage'] = get_storage_class(result['storage'])
        except (ImproperlyConfigured, ImportError):
            # TODO: Log invalid class
            result = {'name': None, 'storage': None}
    return result


def open_stored_file(value, url):
    """
    Deserialize value for a given upload url and return open file.
    Returns None if deserialization fails.
    """
    upload = None
    result = deserialize_upload(value, url)
    filename = result['name']
    storage_class = result['storage']
    if storage_class and filename:
        storage = storage_class()
        if storage.exists(filename):
            upload = storage.open(filename)
            upload.name = os.path.basename(filename)
    return upload


"""Utility module for compatibility between Django and Python versions."""
from __future__ import unicode_literals

from django.conf import settings
try:
    from django.contrib.auth import get_user_model
except ImportError: # pragma: no cover
    # Django < 1.5
    from django.contrib.auth.models import User
    User.USERNAME_FIELD = 'username'
    get_user_model = lambda: User


from __future__ import unicode_literals

import django
from django.core.files import File
from django.core.files.storage import FileSystemStorage
from django.core.files.uploadedfile import SimpleUploadedFile
from django.core.urlresolvers import reverse
from django.test import SimpleTestCase
from django.utils.encoding import python_2_unicode_compatible

from ..utils import serialize_upload
from ..widgets import StickyUploadWidget
from .base import TempFileMixin


@python_2_unicode_compatible
class FakeFieldFile(object):
    """
    Quacks like a FieldFile (has a .url and unicode representation), but
    doesn't require us to care about storages etc.
    """

    url = 'something'

    def __str__(self):
        return self.url


class StickyUploadWidgetTestCase(TempFileMixin, SimpleTestCase):
    """Customized file widget which restored values from the serialized store."""

    def setUp(self):
        super(StickyUploadWidgetTestCase, self).setUp()
        self.widget = StickyUploadWidget()
        self.url = reverse('sticky-upload-default')

    def test_render(self):
        """Default render of the widget without any value."""
        self.assertHTMLEqual(self.widget.render('myfile', None),
            '<input type="file" name="myfile" data-upload-url="/sticky-uploads/default/" />' +
            '<input type="hidden" name="_myfile" />')

    def test_render_with_initial(self):
        """Render with standard FieldFile."""
        value = FakeFieldFile()
        self.assertHTMLEqual(self.widget.render('myfile', value),
            'Currently: <a href="something">something</a> ' +
            '<input id="myfile-clear_id" name="myfile-clear" type="checkbox" />' +
            '<label for="myfile-clear_id"> Clear</label><br />' +
            'Change:<input type="file" name="myfile" data-upload-url="/sticky-uploads/default/" />' +
            '<input type="hidden" name="_myfile" />')

    def test_render_with_restored_file(self):
        """Render with File which has been restored."""
        with open(self.temp_name) as temp:
            currently_html = 'Currently: {0} '.format(self.temp_name)
            #Django does not allow overriding url_markup_template before 1.6
            # remove when 1.4, 1.5 support is dropped
            if django.VERSION < (1, 6):
                currently_html = 'Currently: <a href="#">{0}</a> '.format(self.temp_name)
            value = File(temp)
            setattr(value, '_seralized_location', '1234')
            self.assertHTMLEqual(self.widget.render('myfile', value),
                currently_html +
                '<input id="myfile-clear_id" name="myfile-clear" type="checkbox" />' +
                '<label for="myfile-clear_id"> Clear</label><br />' +
                'Change:<input type="file" name="myfile" data-upload-url="/sticky-uploads/default/" />' +
                '<input type="hidden" name="_myfile" value="1234" />')

    def test_value_from_files(self):
        """Get uploaded file from the FILES as normal."""
        f = SimpleUploadedFile('something.txt', b'content')
        value = self.widget.value_from_datadict(data={}, files={'myfile': f}, name='myfile')
        self.assertEqual(value, f)

    def test_value_from_post(self):
        """Get uploaded file by restoring value from the POST."""
        with self.settings(MEDIA_ROOT=self.temp_dir):
            storage = FileSystemStorage()
            stored = serialize_upload(self.temp_name, storage, self.url)
            value = self.widget.value_from_datadict(data={'_myfile': stored}, files={}, name='myfile')
            self.assertTrue(isinstance(value, File))
            self.assertEqual(value._seralized_location, stored)

    def test_value_from_wrong_urls(self):
        """Widget should not restore a value from a url other than the one specified."""
        with self.settings(MEDIA_ROOT=self.temp_dir):
            storage = FileSystemStorage()
            stored = serialize_upload(self.temp_name, storage, '/some-other/url/')
            value = self.widget.value_from_datadict(data={'_myfile': stored}, files={}, name='myfile')
            self.assertIsNone(value)



from __future__ import unicode_literals

import os
import shutil
import tempfile

from django.conf import settings
from django.core import signing
from django.core.files import File
from django.core.files.storage import FileSystemStorage, DefaultStorage
from django.test import SimpleTestCase
from django.utils import unittest

from .. import utils
from .base import TempFileMixin


class SerializeTestCase(unittest.TestCase):
    """Serialize a file along with its storage."""

    def test_serialize(self):
        """Serialize mapping of file and storage."""
        storage = FileSystemStorage()
        result = utils.serialize_upload('test.png', storage, '/upload/default/')
        expected = signing.dumps({
            'name': 'test.png',
            'storage': 'django.core.files.storage.FileSystemStorage',
        }, salt='/upload/default/')
        self.assertEqual(result, expected)

    def test_lazy_storge(self):
        """Serialize lazy storage such as DefaultStorage."""
        storage = DefaultStorage()
        result = utils.serialize_upload('test.png', storage, '/upload/default/')
        expected = signing.dumps({
            'name': 'test.png',
            'storage': settings.DEFAULT_FILE_STORAGE,
        }, salt='/upload/default/')
        self.assertEqual(result, expected)


class DeserializeTestCase(SimpleTestCase):
    """Deserialize a file along with its storage class."""

    def test_deserialize(self):
        """Deserialize mapping of file and storage."""
        storage = FileSystemStorage()
        value = utils.serialize_upload('test.png', storage, '/upload/default/')
        result = utils.deserialize_upload(value, '/upload/default/')
        expected = {
            'name': 'test.png',
            'storage': FileSystemStorage,
        }
        self.assertEqual(result, expected)

    def test_bad_signature(self):
        """Attempt to restore when SECRET_KEY has changed."""
        storage = FileSystemStorage()
        value = utils.serialize_upload('test.png', storage, '/upload/default/')
        with self.settings(SECRET_KEY='1234'):
            result = utils.deserialize_upload(value, '/upload/default/')
        expected = {
            'name': None,
            'storage': None,
        }
        self.assertEqual(result, expected)

    def test_unknown_storage(self):
        """Attempt to restore storage class which is no longer importable."""
        value = signing.dumps({
            'name': 'test.png',
            'storage': 'does.not.exist',
        }, salt='/upload/default/')
        result = utils.deserialize_upload(value, '/upload/default/')
        expected = {
            'name': None,
            'storage': None,
        }
        self.assertEqual(result, expected)

    def test_incorrect_url(self):
        """Attempt to restore a serialized value generated by a different url."""
        storage = FileSystemStorage()
        value = utils.serialize_upload('test.png', storage, '/upload/default/')
        result = utils.deserialize_upload(value, '/upload/custom/')
        expected = {
            'name': None,
            'storage': None,
        }
        self.assertEqual(result, expected)


class OpenStoredFileTestCase(TempFileMixin, SimpleTestCase):
    """Deserialize and open file from a storage."""

    def test_open_file(self):
        """Restore and open file from storage."""
        with self.settings(MEDIA_ROOT=self.temp_dir):
            storage = FileSystemStorage()
            value = utils.serialize_upload(self.temp_name, storage, '/upload/default/')
            result = utils.open_stored_file(value, '/upload/default/')
            self.assertTrue(isinstance(result, File))
            self.assertEqual(result.name, os.path.basename(self.temp_name))

    def test_bad_signature(self):
        """Attempt to open file when SECRET_KEY has changed."""
        with self.settings(MEDIA_ROOT=self.temp_dir):
            storage = FileSystemStorage()
            value = utils.serialize_upload(self.temp_name, storage, '/upload/default/')
            with self.settings(SECRET_KEY='1234'):
                result = utils.open_stored_file(value, '/upload/default/')
                self.assertIsNone(result)

    def test_unknown_storage(self):
        """Attempt to open file with storage class which is no longer importable."""
        value = signing.dumps({
            'name': self.temp_name,
            'storage': 'does.not.exist',
        }, salt='/upload/default/')
        result = utils.open_stored_file(value, '/upload/default/')
        self.assertIsNone(result)

    def test_file_does_not_exist(self):
        """Restore file not found in the storage."""
        with self.settings(MEDIA_ROOT=self.temp_dir):
            storage = FileSystemStorage()
            value = utils.serialize_upload('test.png', storage, '/upload/default/')
            result = utils.open_stored_file(value, '/upload/default/')
            self.assertIsNone(result)

    def test_incorrect_url(self):
        """Attempt to open file generated by a different url."""
        with self.settings(MEDIA_ROOT=self.temp_dir):
            storage = FileSystemStorage()
            value = utils.serialize_upload(self.temp_name, storage, '/upload/default/')
            result = utils.open_stored_file(value, '/upload/custom/')
            self.assertIsNone(result)


from django.db import models

from .base import mockstorage

class Example(models.Model):
    name = models.CharField(max_length=100)
    upload = models.FileField(storage=mockstorage, upload_to='test/')


from .test_forms import UploadFormTestCase
from .test_integration import FormIntegrationTestCase, ModelFormIntegrationTestCase
from .test_utils import SerializeTestCase, DeserializeTestCase, OpenStoredFileTestCase
from .test_views import UploadViewTestCase
from .test_widgets import StickyUploadWidgetTestCase


from __future__ import unicode_literals

from django import forms
from django.core.urlresolvers import reverse_lazy
from django.utils.safestring import mark_safe

from . import __version__
from .utils import open_stored_file


class StickyUploadWidget(forms.ClearableFileInput):
    """Customize file uploader widget to handle AJAX upload and preserve value."""
    # template to use when there is no url
    url_markup_template_tmp = '{1}'

    class Media(object):
        js = (
            'stickyuploads/js/django-uploader.bundle.min.js?v=%s' % __version__,
        )

    def __init__(self, *args, **kwargs):
        self.url = kwargs.pop('url', reverse_lazy('sticky-upload-default'))
        super(StickyUploadWidget, self).__init__(*args, **kwargs)

    def get_hidden_name(self, name):
        """Get expected hidden name from the original field name."""
        return '_' + name

    def value_from_datadict(self, data, files, name):
        """Returns uploaded file from serialized value."""
        upload = super(StickyUploadWidget, self).value_from_datadict(data, files, name)
        if upload is not None:
            # File was posted or cleared as normal
            return upload
        else:
            # Try the hidden input
            hidden_name = self.get_hidden_name(name)
            value = data.get(hidden_name, None)
            if value is not None:
                upload = open_stored_file(value, self.url)
                if upload is not None:
                    setattr(upload, '_seralized_location', value)
        return upload

    def render(self, name, value, attrs=None):
        """Include a hidden input to stored the serialized upload value."""
        location = getattr(value, '_seralized_location', '')
        if location and not hasattr(value, 'url'):
            value.url = '#'
            if hasattr(self, 'get_template_substitution_values'):
                # Django 1.8+
                self.template_with_initial = (
                    '%(initial_text)s: %(initial)s %(clear_template)s'
                    '<br />%(input_text)s: %(input)s')
            elif hasattr(self, 'url_markup_template'):
                # Django 1.6-1.7
                self.url_markup_template = self.url_markup_template_tmp               
        attrs = attrs or {}
        attrs.update({'data-upload-url': self.url})
        parent = super(StickyUploadWidget, self).render(name, value, attrs=attrs)
        hidden_name = self.get_hidden_name(name)
        hidden = forms.HiddenInput().render(hidden_name, location)
        return mark_safe(parent + hidden)



"""Test integration of StickyUploadWidget with full Forms and ModelForms."""
from __future__ import unicode_literals

from django import forms

from django.core.files.storage import FileSystemStorage
from django.core.files.uploadedfile import SimpleUploadedFile
from django.core.urlresolvers import reverse
from django.test import TestCase, SimpleTestCase

from ..utils import serialize_upload
from ..widgets import StickyUploadWidget
from . import Example
from .base import TempFileMixin, mockstorage


class TestForm(forms.Form):
    name = forms.CharField()
    upload = forms.FileField(widget=StickyUploadWidget)


class FormIntegrationMixin(object):
    """Tests for form compatibility."""

    form_class = None
    url_conf = 'stickyuploads.tests.urls'

    def setUp(self):
        super(FormIntegrationMixin, self).setUp()
        self.url = reverse('sticky-upload-default')

    def test_valid_from_files(self):
        """Fallback functionality to get file from passed FILES."""
        data = {'name': 'foo'}
        files = {'upload': SimpleUploadedFile('something.txt', b'content')}
        form = self.form_class(data=data, files=files)
        self.assertTrue(form.is_valid())

    def test_valid_from_post(self):
        """Restore file from serialized value."""
        with self.settings(MEDIA_ROOT=self.temp_dir):
            storage = FileSystemStorage()
            stored = serialize_upload(self.temp_name, storage, self.url)
            data = {'name': 'foo', '_upload': stored}
            form = self.form_class(data=data, files={})
            self.assertTrue(form.is_valid())

    def test_keep_value_on_failure(self):
        """Hidden input should keep serialized value when the form is not valid."""
        with self.settings(MEDIA_ROOT=self.temp_dir):
            storage = FileSystemStorage()
            stored = serialize_upload(self.temp_name, storage, self.url)
            data = {'name': '', '_upload': stored}
            form = self.form_class(data=data, files={})
            self.assertFalse(form.is_valid())
            expected = 'value="{0}"'.format(stored)
            self.assertIn(stored, form.as_p())


class FormIntegrationTestCase(TempFileMixin, FormIntegrationMixin, SimpleTestCase):
    """Using StickyUploadWidget in a Form class."""

    form_class = TestForm


class ExampleForm(forms.ModelForm):

    class Meta:
        model = Example
        fields = ('name', 'upload', )
        widgets = {
            'upload': StickyUploadWidget
        }


class ModelFormIntegrationTestCase(TempFileMixin, FormIntegrationMixin, SimpleTestCase):
    """Using StickyUploadWidget in a ModelForm class."""

    form_class = ExampleForm

    def test_save(self):
        """Save the model from the valid model form."""
        with self.settings(MEDIA_ROOT=self.temp_dir):
            storage = FileSystemStorage()
            stored = serialize_upload(self.temp_name, storage, self.url)
            data = {'name': 'foo', '_upload': stored}
            form = self.form_class(data=data, files={})
            self.assertTrue(form.is_valid())
            instance = form.save()
            self.assertTrue(instance.upload)


import tempfile

from django.core.files.storage import FileSystemStorage


class TempFileSystemStorage(FileSystemStorage):
    """Storage class to store files in the system /tmp."""

    def __init__(self):
        super(TempFileSystemStorage, self).__init__(
            location=tempfile.gettempdir(), base_url=None)

    def url(self, name):
        """Files are not accessable via url."""
        raise NotImplementedError()


"""Utitilies for writing tests manipulating temporary files/directories."""

import shutil
import tempfile

from mock import Mock


class MockStorage(Mock):

    def save(self, name, upload):
        return name

    def url(self, name):
        return '/uploads/{0}'.format(name)

    def get_valid_name(self, name):
        return name


mockstorage = MockStorage()


class TempFileMixin(object):

    def setUp(self):
        super(TempFileMixin, self).setUp()
        self.temp_dir = tempfile.mkdtemp()
        _, self.temp_name = tempfile.mkstemp(dir=self.temp_dir)
        with open(self.temp_name, 'w') as f:
            f.write('X')

    def tearDown(self):
        super(TempFileMixin, self).tearDown()
        shutil.rmtree(self.temp_dir, ignore_errors=True)


from __future__ import unicode_literals

import os
import json
import shutil
import tempfile

from django.core.urlresolvers import reverse
from django.test import TestCase

from ..compat import get_user_model
from ..storage import TempFileSystemStorage
from ..utils import serialize_upload
from .base import TempFileMixin


class UploadViewTestCase(TempFileMixin, TestCase):
    """View to handle background AJAX upload."""

    url_conf = 'stickyuploads.tests.urls'

    def setUp(self):
        super(UploadViewTestCase, self).setUp()
        User = get_user_model()
        self.user = User.objects.create_user(**{User.USERNAME_FIELD: 'test', 'password': 'test'})
        self.client.login(username='test', password='test')
        self.url = reverse('sticky-upload-default')

    def test_upload(self):
        """Handle a new file upload."""
        with open(self.temp_name, 'rb') as f:
            data = {'upload': f}
            response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 200)
        result = json.loads(response.content.decode('utf-8'))
        storage = TempFileSystemStorage()
        filename = os.path.basename(self.temp_name)
        expected = {
            'is_valid': True,
            'filename': filename,
            'url': None,
            'stored': serialize_upload(filename, storage, self.url),
        }
        self.assertEqual(result, expected)

    def test_upload_not_allowed(self):
        """User must be authenticated for the default upload."""
        self.client.logout()
        with open(self.temp_name, 'rb') as f:
            data = {'upload': f}
            response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 403)

    def test_empty_file(self):
        """Uploading an empty file is not valid."""
        with open(self.temp_name, 'w') as f:
            f.write('')
        with open(self.temp_name, 'rb') as f:
            data = {'upload': f}
            response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 200)
        result = json.loads(response.content.decode('utf-8'))
        self.assertEqual(result, {'is_valid': False})


from __future__ import unicode_literals

import json

from django.core.files.storage import get_storage_class
from django.http import HttpResponse, HttpResponseForbidden
from django.views.generic import View

from .forms import UploadForm


class UploadView(View):
    """Base view class for accepting file uploads."""

    form_class = UploadForm
    storage_class = 'stickyuploads.storage.TempFileSystemStorage'

    def post(self, *args, **kwargs):
        """Save file and return saved info or report errors."""
        if self.upload_allowed():
            form = self.get_upload_form()
            result = {}
            if form.is_valid():
                storage = self.get_storage()
                result['is_valid'] = True
                info = form.stash(storage, self.request.path)
                result.update(info)
            else:
                result['is_valid'] = False
            return HttpResponse(json.dumps(result), content_type='application/json')
        else:
            return HttpResponseForbidden()

    def get_storage(self):
        """Get storage instance for saving the temporary upload."""
        return get_storage_class(self.storage_class)()

    def upload_allowed(self):
        """Check if the current request is allowed to upload files."""
        return self.request.user.is_authenticated()

    def get_upload_form(self):
        """Construct form for accepting file upload."""
        return self.form_class(self.request.POST, self.request.FILES)